# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gitlabracadabra.dictutils import update_dict_with_defaults
from gitlabracadabra.tests.case import TestCase


class TestDictUtils(TestCase):
    """Test gitlabracadabra.dictutils."""

    def test_update_dict_with_defaults(self):
        """Test update_dict_with_defaults."""
        defaults = {
            '.foo': {
                'not_overridden_string': 'foo',
                'not_overridden_dict': {'foo': {'bar': 'buz'}},
                'not_overridden_list': ['ab', 'cd'],
                'overridden_string': 'D_foo',
                'overridden_dict': {'D_foo': {'D_bar': 'D_buz'}},
                'overridden_list': ['D_ab', 'D_cd'],
            },
        }
        target = {
            '.foo': {
                'overridden_string': 'O_foo',
                'overridden_dict': {'D_foo': {'O_bar': 'O_buz'}},
                'overridden_list': ['O_ab', 'O_cd'],
            },
        }
        update_dict_with_defaults(target, defaults)
        expected = {
            '.foo': {
                'not_overridden_string': 'foo',
                'not_overridden_dict': {'foo': {'bar': 'buz'}},
                'not_overridden_list': ['ab', 'cd'],
                'overridden_string': 'O_foo',
                'overridden_dict': {'D_foo': {'D_bar': 'D_buz', 'O_bar': 'O_buz'}},
                'overridden_list': ['O_ab', 'O_cd'],
            },
        }
        self.assertEqual(target, expected)

    def test_update_dict_with_defaults_aggregate(self):
        """Test update_dict_with_defaults with aggregate=True."""
        defaults = {
            '.foo': {
                'not_overridden_string': 'foo',
                'not_overridden_dict': {'foo': {'bar': 'buz'}},
                'not_overridden_list': ['ab', 'cd'],
                'overridden_string': 'D_foo',
                'overridden_dict': {'D_foo': {'D_bar': 'D_buz'}},
                'overridden_list': ['D_ab', 'D_cd'],
            },
        }
        target = {
            '.foo': {
                'overridden_string': 'O_foo',
                'overridden_dict': {'D_foo': {'O_bar': 'O_buz'}},
                'overridden_list': ['O_ab', 'O_cd'],
            },
        }
        update_dict_with_defaults(target, defaults, aggregate=True)
        expected = {
            '.foo': {
                'not_overridden_string': 'foo',
                'not_overridden_dict': {'foo': {'bar': 'buz'}},
                'not_overridden_list': ['ab', 'cd'],
                'overridden_string': 'O_foo',
                'overridden_dict': {'D_foo': {'D_bar': 'D_buz', 'O_bar': 'O_buz'}},
                'overridden_list': ['D_ab', 'D_cd', 'O_ab', 'O_cd'],
            },
        }
        self.assertEqual(target, expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest.mock import call, patch

from gitlabracadabra.objects.project import GitLabracadabraProject
from gitlabracadabra.tests import my_vcr
from gitlabracadabra.tests.case import TestCaseWithManager


class TestProjectRenameBranches(TestCaseWithManager):
    @my_vcr.use_cassette
    def test_rename_branches(self, cass):
        obj = GitLabracadabraProject('memory', 'test/test_rename_branches', {
            'rename_branches': [
                {'old_name': 'new_name'},
                {'branch2': 'branch3'},
                {'branch1': 'branch2'},
                {'does_not_exists': 'already_exists'},
                {'does_not_exists': 'should_print_error'},
            ],
        })
        self.assertEqual(obj.errors(), [])
        with patch('gitlabracadabra.mixins.rename_branches.logger', autospec=True) as logger:
            obj.process()
        logger.assert_has_calls([
            call.info('[%s] Renaming branch from %s to %s', 'test/test_rename_branches', 'old_name', 'new_name'),
            call.info('[%s] Renaming branch from %s to %s', 'test/test_rename_branches', 'branch2', 'branch3'),
            call.info('[%s] Renaming branch from %s to %s', 'test/test_rename_branches', 'branch1', 'branch2'),
            call.info('[%s] NOT Renaming branch from %s to %s (old branch not found)',
                      'test/test_rename_branches', 'does_not_exists', 'should_print_error'),
        ])
        self.assertTrue(cass.all_played)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

from setuptools import find_packages, setup

import gitlabracadabra


with open('README.md', 'r') as readme_file:
    readme = readme_file.read()

setup(
    name='gitlabracadabra',
    version=gitlabracadabra.__version__,
    description='Adds some magic to GitLab',
    long_description=readme,
    long_description_content_type='text/markdown',
    author=gitlabracadabra.__author__,
    author_email=gitlabracadabra.__email__,
    license=gitlabracadabra.__license__,
    url='https://gitlab.com/gitlabracadabra/gitlabracadabra',
    packages=find_packages(),
    package_data={
        'gitlabracadabra.tests.unit': [
            'fixtures/*',
            'fixtures/testrepo.git/*',
            'fixtures/testrepo.git/objects/*/*',
            'fixtures/testrepo.git/refs/*/*',
        ],
        'gitlabracadabra.tests': [
            'python-gitlab.cfg',
        ],
    },
    install_requires=[
        'jsonschema>=2.6.0',
        'python-gitlab>=1.6.0',
        'PyYAML',
        'pygit2',
        'semantic_version',
    ],
    entry_points={
        'console_scripts': [
            'gitlabracadabra = gitlabracadabra.cli:main',
        ],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='gitlab api yaml',
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from copy import deepcopy

from gitlab.exceptions import GitlabDeleteError

from gitlabracadabra.gitlab.access_levels import access_level


logger = logging.getLogger(__name__)


class MembersMixin(object):
    """Object with members."""

    """_process_members()

    Process the members param.
    """
    def _process_members(self, param_name, param_value, dry_run=False, skip_save=False):
        assert param_name == 'members'  # noqa: S101
        assert not skip_save  # noqa: S101
        param_value = deepcopy(param_value)
        unknown_members = self._content.get('unknown_members', 'warn')
        current_members = dict([[member.username, member] for member in self._obj.members.list(all=True)])
        # We first check for already existing members
        for _member_username, member in sorted(current_members.items()):
            self.connection.user_cache.map_user(member.id, member.username)
            if member.username in param_value.keys():
                target_access_level = access_level(param_value[member.username])
                if member.access_level != target_access_level:
                    if dry_run:
                        logger.info('[%s] NOT Changing user %s (%s) access level: %s -> %s (dry-run)',
                                    self._name, member.username, member.name,
                                    member.access_level, target_access_level)
                    else:
                        logger.info('[%s] Changing user %s (%s) access level: %s -> %s',
                                    self._name, member.username, member.name,
                                    member.access_level, target_access_level)
                        member.access_level = target_access_level
                        member.save()
                param_value.pop(member.username)
            else:
                if unknown_members in ['delete', 'remove']:
                    if dry_run:
                        logger.info('[%s] NOT Removing member %s (%s) (dry-run)',
                                    self._name, member.username, member.name)
                    else:
                        logger.info('[%s] Removing member %s (%s)',
                                    self._name, member.username, member.name)
                        try:
                            member.delete()
                        except GitlabDeleteError as e:
                            logger.warning('[%s] Unable to remove member %s (%s): %s',
                                           self._name, member.username, member.name, e.error_message)
                elif unknown_members not in ['ignore', 'skip']:
                    logger.warning('[%s] NOT Removing member: %s (%s)',
                                   self._name, member.username, member.name)
        # Remaining members
        for target_username, target_user_access in sorted(param_value.items()):
            user_id = self.connection.user_cache.id_from_username(target_username)
            if user_id is None:
                logger.warning('[%s] User not found %s',
                               self._name, target_username)
                continue
            target_access_level = access_level(target_user_access)
            if dry_run:
                logger.info('[%s] NOT Adding user %s: %s -> %s (dry-run)',
                            self._name, target_username, 0, target_access_level)
            else:
                logger.info('[%s] Adding user %s: %s -> %s',
                            self._name, target_username, 0, target_access_level)
                self._obj.members.create({
                    'user_id': user_id,
                    'access_level': target_access_level,
                })



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from logging import getLogger
from re import IGNORECASE, Pattern, compile as re_compile, error as re_error
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from typing import Callable, List, Match, Optional, Union

    InputData = Union[
        List[str],
        Callable[[], List[str]],
    ]
    Patterns = List[Union[str, Pattern[str]]]


logger = getLogger(__name__)


class Matcher(object):
    """Matcher."""

    def __init__(
        self,
        patterns: Union[str, List[str]],
        *,
        log_prefix: str = '',
    ) -> None:
        """Initialize a matcher.

        Args:
            patterns: A pattern or list of patterns.
            log_prefix: Log prefix.
        """
        self._log_prefix = log_prefix
        if not isinstance(patterns, list):
            patterns = [patterns]
        self._patterns: Patterns = []
        for pattern in patterns:
            parsed = self._parse_pattern(pattern)
            if parsed is not None:
                self._patterns.append(parsed)

    def match(
        self,
        input_data: InputData,
    ) -> List[Match]:
        """Filer.

        Args:
            input_data: Either a list of string or an input function, called only when needed (and at most once).

        Returns:
            List of matches.
        """
        has_regex = any(isinstance(pattern, Pattern) for pattern in self._patterns)
        if not isinstance(input_data, list) and has_regex:
            input_data = input_data()
        if isinstance(input_data, list):
            return self._match_list(input_data)
        return self._match_all()

    def _match_list(self, input_data: List[str]) -> List[Match[str]]:
        matched_items: List[Match[str]] = []
        for current_item in input_data:
            match = self._match_item(current_item)
            if match:
                matched_items.append(match)
        return matched_items

    def _match_item(self, current_item: str) -> Optional[Match[str]]:
        if current_item in self._patterns:
            return re_compile('^.*$').match(current_item)
        for pattern in self._patterns:
            if isinstance(pattern, Pattern):
                match = pattern.match(current_item)
                if match:
                    return match
        return None

    def _match_all(self) -> List[Match[str]]:
        matched_items: List[Match[str]] = []
        for pattern in self._patterns:
            if not isinstance(pattern, Pattern):
                match = re_compile('^.*$').match(pattern)
                if match:
                    matched_items.append(match)
        return matched_items

    def _parse_pattern(self, pattern: str) -> Union[None, str, Pattern[str]]:
        """Parse a pattern.

        Args:
            pattern: The pattern as string.

        Returns:
            A string for exact match or a pattern.
        """
        if pattern.startswith('/'):
            flags_str = pattern.rsplit('/', 1).pop()
            flags = 0
            for flag in flags_str:
                if flag == 'i':
                    flags |= IGNORECASE
                else:
                    logger.warning(
                        '%sInvalid regular expression flag %s in %s. Flag ignored.',
                        self._log_prefix,
                        flag,
                        pattern,
                    )
            try:
                return re_compile(
                    '^{0}$'.format(pattern[1:pattern.rindex('/')]),
                    flags,
                )
            except re_error as err:
                logger.warning(  # noqa: G200
                    '%sInvalid regular expression %s: %s. Skipping pattern.',
                    self._log_prefix,
                    pattern,
                    str(err),
                )
                return None
        return pattern



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp
from unittest import TestCase as BaseTestCase
from unittest.mock import MagicMock, patch

from gitlabracadabra.containers.registries import Registries
from gitlabracadabra.gitlab.connections import GitlabConnections


class TestCase(BaseTestCase):
    """TestCase."""

    def setUp(self) -> None:
        """Test setup.

        Create temporary directory, and mock user_cache_dir_path().
        """
        super().setUp()

        Registries().reset()

        self._temp_dir = Path(mkdtemp())
        self._user_cache_dir_path_mock = MagicMock()
        self._user_cache_dir_path_mock.return_value = self._temp_dir / 'cache'
        self._user_cache_dir_path_patch = patch(
            'gitlabracadabra.disk_cache.user_cache_dir_path',
            self._user_cache_dir_path_mock,
        )
        self._user_cache_dir_path_patch.start()

    def tearDown(self) -> None:
        """Test teardown.

        Delete temporary directory, and mock user_cache_dir_path().
        """
        self._user_cache_dir_path_patch.stop()
        rmtree(self._temp_dir)
        super().tearDown()


class TestCaseWithManager(TestCase):
    """TestCase with GitLab connection."""

    def setUp(self) -> None:
        """Test setup.

        Inject a GitLab connection.
        """
        super().setUp()
        pygitlab_config = str((Path(__file__).parent / 'python-gitlab.cfg').absolute())
        GitlabConnections().load(default_id='localhost', config_files=[pygitlab_config], debug=False)
        GitlabConnections().get_connection(None, auth=False)

    def tearDown(self) -> None:
        """Test teardown.

        Prune the GitLab connections.
        """
        GitlabConnections().load(default_id=None, config_files=None, debug=False)
        super().tearDown()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 Mathieu Parent <math.parent@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gitlabracadabra.objects.application_settings import GitLabracadabraApplicationSettings
from gitlabracadabra.tests import my_vcr
from gitlabracadabra.tests.case import TestCaseWithManager


class TestApplicationSettings(TestCaseWithManager):
    @my_vcr.use_cassette
    def test_settings(self, cass):
        obj = GitLabracadabraApplicationSettings('memory', 'application_settings', {
            'admin_notification_email': 'admin@example.org',
            'after_sign_out_path': 'https://logout.example.org/',
            'after_sign_up_text': '/sign/up',
            'akismet_api_key': 'abcd',
            'akismet_enabled': True,
            'allow_local_requests_from_hooks_and_services': True,
            'archive_builds_in_human_readable': '1 month',
            'authorized_keys_enabled': False,
            'auto_devops_domain': 'app.example.org',
            'auto_devops_enabled': False,
            'commit_email_hostname': 'users.noreply.gitlab.example.org',
            'container_registry_token_expire_delay': 10,
            'default_artifacts_expire_in': '15 days',
            'default_branch_protection': 1,
            # 'default_ci_config_path': 'global/path', # TODO
            'default_group_visibility': 'internal',
            'default_project_creation': 1,
            'default_projects_limit': 42,
            'default_project_visibility': 'internal',
            'default_snippet_visibility': 'internal',
            'diff_max_patch_bytes': 102420,
            'disabled_oauth_sign_in_sources': ['github'],
            'dns_rebinding_protection_enabled': False,
            'domain_blacklist_enabled': True,
            'domain_blacklist': ['foo.example.com'],
            'domain_whitelist': ['bar.example.com'],
            'dsa_key_restriction': 1024,
            'ecdsa_key_restriction': 384,
            'ed25519_key_restriction': 256,
            'email_author_in_body': True,
            'enabled_git_access_protocol': 'http',
            'enforce_terms': True,
            # 'external_auth_client_cert': 'FOO',  # FIXME 'is not a valid X509 certificate.'
            # 'external_auth_client_key': 'BAR', # 'could not read private key, is the passphrase correct?'
            'external_auth_client_key_pass': 'BAZ',
            'external_authorization_service_default_label': 'toto',
            # 'external_authorization_service_enabled': True,  # FIXME Or the API will fail
            'external_authorization_service_timeout': 0.42,
            'external_authorization_service_url': 'https://authz.example.org/',
            'first_day_of_week': 'sunday',
            'gitaly_timeout_default': 110,
            'gitaly_timeout_fast': 20,
            'gitaly_timeout_medium': 60,
            'grafana_enabled': True,
            'grafana_url': '/-/grafana2',
            'gravatar_enabled': False,
            'hashed_storage_enabled': True,
            'help_page_hide_commercial_content': True,
            'help_page_support_url': 'https://help.example.org/',
            'help_page_text': 'Help!',
            'hide_third_party_offers': True,
            'home_page_url': 'https://home.example.org/',
            'housekeeping_bitmaps_enabled': False,
            'housekeeping_enabled': False,
            'housekeeping_full_repack_period': 100,
            'housekeeping_gc_period': 400,
            'housekeeping_incremental_repack_period': 20,
            'html_emails_enabled': False,
            'import_sources': [
                'github',
                'bitbucket',
                # 'bitbucket_server',
                'gitlab',
                'google_code',
                'fogbugz',
                'git',
                'gitlab_project',
                # 'gitea',
                'manifest',
                # 'phabricator',
            ],
            'instance_statistics_visibility_private': True,
            'local_markdown_version': 1,
            'max_artifacts_size': 200,
            'max_attachment_size': 20,
            'max_pages_size': 200,
            'metrics_enabled': True,
            'metrics_host': 'influxdb.example.org',
            'metrics_method_call_threshold': 20,
            'metrics_packet_size': 2,
            'metrics_pool_size': 32,
            'metrics_port': 18089,
            'metrics_sample_interval': 30,
            'metrics_timeout': 20,
            'mirror_available': False,
            'pages_domain_verification_enabled': False,
            'password_authentication_enabled_for_git': False,
            'password_authentication_enabled_for_web': False,
            'performance_bar_allowed_group_path': 'test',
            'plantuml_enabled': True,
            'plantuml_url': 'https://plantuml.example.org/',
            'polling_interval_multiplier': 2.0,
            'project_export_enabled': False,
            'prometheus_metrics_enabled': False,
            'protected_ci_variables': True,
            'recaptcha_enabled': True,
            'recaptcha_private_key': 'FOO',
            'recaptcha_site_key': 'BAR',
            'receive_max_input_size': 42,
            'repository_checks_enabled': False,
            # 'repository_storages': ['nfs'],  # FIXME
            'require_two_factor_authentication': True,
            'restricted_visibility_levels': ['public'],
            'rsa_key_restriction': 1024,
            'send_user_confirmation_email': True,
            'session_expire_delay': 20160,
            'shared_runners_enabled': False,
            'shared_runners_text': 'Shared runner text',
            'sign_in_text': 'Sign in text',
            'signup_enabled': False,
            'terminal_max_session_time': 3600,
            'terms': 'Terms',
            'throttle_authenticated_api_enabled': True,
            'throttle_authenticated_api_period_in_seconds': 7200,
            'throttle_authenticated_api_requests_per_period': 14400,
            'throttle_authenticated_web_enabled': True,
            'throttle_authenticated_web_period_in_seconds': 7200,
            'throttle_authenticated_web_requests_per_period': 14400,
            'throttle_unauthenticated_enabled': True,
            'throttle_unauthenticated_period_in_seconds': 7200,
            'throttle_unauthenticated_requests_per_period': 7200,
            'time_tracking_limit_to_hours': True,
            'two_factor_grace_period': 96,
            'unique_ips_limit_enabled': True,
            'unique_ips_limit_per_user': 20,
            'unique_ips_limit_time_window': 7200,
            'usage_ping_enabled': True,
            'user_default_external': True,
            'user_default_internal_regex': '.*@.*\\.example.org',
            'user_oauth_applications': False,
            'user_show_add_ssh_key_message': False,
            'version_check_enabled': True,
            'web_ide_clientside_preview_enabled': True,
        })
        obj.process()
        self.assertTrue(cass.all_played)
